﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Runtime.InteropServices;
using System.Text;
using System.Threading.Tasks;
using System.Windows.Forms;
using System.Drawing;
using System.Drawing.Drawing2D;
using System.Drawing.Imaging;
using System.Data.SqlTypes;

namespace MariCam
{
    public class Program
    {
        public enum ConnectionType : int
        {
            CSI = 0,
            USB = 1
        }

        public enum FrameType : int
        {
            RAW = 0,
            AMPLITUDE = 1,
            DEPTH = 2
        }

        [StructLayout(LayoutKind.Sequential)]
        public struct FrameFormat
        {
            public ushort Width;

            public ushort Height;

            public int FrameType;

            public ulong Timestamp;
        }

        public const string LibraryPath = "/usr/lib/libArducamDepthCamera2c.so";

        [System.Runtime.InteropServices.DllImport(LibraryPath)]
        public static unsafe extern IntPtr createArducamDepthCamera();

        [System.Runtime.InteropServices.DllImport(LibraryPath)]
        public static unsafe extern IntPtr arducamCameraOpen(IntPtr camera, ConnectionType connectionType, int deviceNumber);

        [System.Runtime.InteropServices.DllImport(LibraryPath)]
        public static unsafe extern IntPtr arducamCameraStart(IntPtr camera, FrameType frameType);

        [System.Runtime.InteropServices.DllImport(LibraryPath)]
        public static unsafe extern IntPtr arducamCameraRequestFrame(IntPtr camera, int timeout);

        [System.Runtime.InteropServices.DllImport(LibraryPath)]
        public static extern IntPtr arducamCameraReleaseFrame(IntPtr camera, IntPtr frameBuffer);

        [System.Runtime.InteropServices.DllImport(LibraryPath)]
        public static extern FrameFormat arducamCameraGetFormat(IntPtr camera, int frameType);

        [System.Runtime.InteropServices.DllImport(LibraryPath)]
        public static extern IntPtr arducamCameraGetDepthData(IntPtr frameBuffer);

        public static unsafe void Main(string[] args)
        {
            Console.WriteLine("Setting up Camera");

            var camera = createArducamDepthCamera();

            arducamCameraOpen(camera, ConnectionType.CSI, 0);

            arducamCameraStart(camera, FrameType.DEPTH);

            var tempBuffer = arducamCameraRequestFrame(camera, 200);

            var tempFormat = arducamCameraGetFormat(tempBuffer, 2);

            var buffer = new Bitmap(tempFormat.Width, tempFormat.Height, PixelFormat.Format32bppArgb);

            arducamCameraReleaseFrame(camera, tempBuffer);

            var gradients = GenerateGradients(new List<Color>()
            {
                Color.FromArgb(255, 64, 64, 128), //Violet
                Color.FromArgb(255, 0, 0, 255), //Blue
                Color.FromArgb(255, 0, 255, 255), //Cyan
                Color.FromArgb(255, 255, 255, 0), //Yellow
                Color.FromArgb(255, 255, 165, 0), //Orange
                Color.FromArgb(255, 255, 0, 0) //Red
            }, 255);

            var form = new Form();

            form.Text = "MariCam";

            form.WindowState = FormWindowState.Maximized;

            form.StartPosition = FormStartPosition.CenterScreen;

            form.Show();

            form.Paint += (object sender, PaintEventArgs e) =>
            {
                var frameBuffer = arducamCameraRequestFrame(camera, 200);

                var format = arducamCameraGetFormat(frameBuffer, 2);

                var graphics = e.Graphics;

                graphics.InterpolationMode = InterpolationMode.Bicubic;

                graphics.Clear(Color.Black);

                float* depthData = (float*)arducamCameraGetDepthData(frameBuffer);

                float min = float.MaxValue;
                float max = float.MinValue;

                for (int y = 0; y < format.Height; y++)
                {
                    for (int x = 0; x < format.Width; x++)
                    {
                        var depth = *(depthData + (y * format.Width) + x);

                        if (depth < min)
                            min = depth;

                        if (depth > max)
                            max = depth;
                    }
                }

                var bufferData = buffer.LockBits(new Rectangle(0, 0, buffer.Width, buffer.Height), ImageLockMode.ReadWrite, buffer.PixelFormat);

                var bufferPixels = (uint*)bufferData.Scan0;

                for (int y = 0; y < format.Height; y++)
                {
                    for (int x = 0; x < format.Width; x++)
                    {
                        var depth = *(depthData + (y * format.Width) + x);

                        //maxNew − minNew / maxOld − minOld * (v − minOld) + minNew
                        var color = gradients[(int)(((float)((gradients.Count - 1) - 0) / (float)(max - min) * (depth - min)) + 0)];

                        *(bufferPixels + ((y * format.Width) + x)) = ((uint)255 << 24) | ((uint)(color.R) << 16) | ((uint)(color.G) << 8) | (byte)(color.B);
                    }
                }

                buffer.UnlockBits(bufferData);

                graphics.DrawImage(buffer, new Rectangle(0, 0, form.ClientSize.Width, form.ClientSize.Height));

                arducamCameraReleaseFrame(camera, frameBuffer);

                form.Invalidate();
            };

            form.Invalidate();

            Application.DoEvents();
        }

        public static List<Color> GenerateGradients(List<Color> colors, int steps)
        {
            var gradients = new List<Color>();

            for (int i = 0; i < colors.Count - 1; i++)
            {
                var from = colors[i];

                var to = colors[i + 1];

                for (int s = 0; s < steps; s++)
                {
                    gradients.Add(Color.FromArgb(
                        255,
                        (byte)(from.R + (s / (float)steps) * (to.R - from.R)),
                        (byte)(from.G + (s / (float)steps) * (to.G - from.G)),
                        (byte)(from.B + (s / (float)steps) * (to.B - from.B))
                    ));
                }
            }

            return gradients;
        }
    }
}
